;;;; Editing improvements

;; fix Emacs' definition of a sentence
(setq sentence-end-double-space nil)

;; the utter basics
(general-define-key
 "C-h"   'backward-delete-char
 "C-w"   'backward-kill-word
 "C-u"   'cp-kill-line-0
 ;; A better solution would be to hack kill-line/kill-visual-line to
 ;; kill the region if a region is active. The flexibility of those
 ;; functions is very desirable - you can delete from point to line
 ;; end, clear a line to enter some other text on it (C-a C-k/"a k"),
 ;; or delete one or more whole lines (C-1 C-k/"1 k" or C-<num>
 ;; C-k/"<num> k").
 ;; "C-y"   'whole-line-or-region-yank
 "C-S-k" 'whole-line-or-region-kill-region
 "C-,"   'backward-paragraph
 "C-."   'forward-paragraph
 "C-<"   'beginning-of-buffer
 "C->"   'end-of-buffer)

(define-key isearch-mode-map (kbd "C-h") 'isearch-delete-char)
(defun cp-kill-line-0 ()
  (interactive)
  (kill-line 0))

(global-unset-key (kbd "C-x C-r"))
(general-define-key
 :prefix "C-x C-r"
 "C-i" 'string-insert-rectangle
 "C-r" 'replace-rectangle
 "C-k" 'kill-rectangle)

(defun cp/downcase-dwim (arg)
  "Like `downcase-word', but if region is active, run
`downcase-region' instead. Unlike `downcase-region', rectangular
regions are handled correctly as well.")

;; open-line should always move to the beginning of the current line
;; first, so one can run it anywhere. I also want it to indent it to
;; the next or the previous line...

;; These affect org-meta-return and org-insert-heading-respect-content
;;

;; (defadvice open-line
;;     (before open-line-bol activate)
;;   (beginning-of-visual-line))
;; (defadvice open-line
;;     (after open-line-indent activate)
;;   (if (not (looking-at "\\** "))
;;       (indent-for-tab-command)))

;; (global-set-key (kbd "C-o") 'open-line)
(defun cp/open-line ()
  "The opposite of `cp/open-line-before'. Start a new line below
the current line, regardless of where point is in the current
line. Will not affect the content of the current line. Applies
the correct indentation according to the mode."
  (interactive)
  (progn
    (end-of-line)
    (electric-indent-just-newline t)
    (indent-according-to-mode)))

(defun cp/open-line-before (arg)
  "Like `open-line' but a bit more sane.

In org-mode, run org-open-line.

In other modes, insert a newline above the current line,
regardless of where point is in current line. Will not affect
content of current line. Applies the correct indentation
according to the mode."
  (interactive "p")
  (if (and (equal major-mode 'org-mode)
           (org-at-table-p))
      (org-open-line arg)
    (progn

      (beginning-of-line)
      (electric-indent-just-newline t)
      (forward-char -1)
      (indent-according-to-mode))))
(global-set-key (kbd "C-o") 'cp/open-line-before)

;; I wish you could press e.g. S-o in god-mode to get C-S-o.
;; 10/10 would make life better.
(global-set-key (kbd "C-S-o") 'cp/open-line)

(setq dired-bind-jump nil)
(global-set-key (kbd "C-x C-j") 'join-line)

;; C-z (suspend-frame) is utterly useless and disruptive - good place to put universal-argument
(global-unset-key (kbd "C-z"))
(global-set-key (kbd "C-z") 'universal-argument)
(global-set-key (kbd "C-x C-z") 'repeat)
(global-set-key (kbd "C-x C-;") 'repeat-complex-command)

;; (defun cp-kill-line (&optional arg)
;;   (interactive)
;;   (if arg
;;       (if mark-active
;;           (kill-region arg)
;;         (kill-line arg))
;;     (if mark-active
;;       (kill-region)
;;       (kill-line))))
;; (global-set-key (kbd "C-k") 'cp-kill-line)
;; Chris Done (god-mode) recommends this, but Emacs' repeat command is almost useless...
;; (global-set-key (kbd "C-.") 'repeat)

;;;; I like the transpose commands
(global-unset-key (kbd "C-t"))
(general-define-key
 :prefix "C-t"
 "C-t C-c" 'transpose-chars
 "C-t C-w" 'transpose-words
 "C-t C-l" 'transpose-lines
 "C-t C-s" 'transpose-sentences
 "C-t C-e" 'transpose-sexps
 "C-t C-p" 'transpose-paragraphs)

(defun cp/delete-trailing-whitespace ()
  (unless (derived-mode-p 'markdown-mode)
    (delete-trailing-whitespace)))

(add-hook 'before-save-hook 'cp/delete-trailing-whitespace)

(require 'flex-isearch)
(flex-isearch-mode 1)

(general-define-key
 "C-M-s" 'flex-isearch-forward
 "C-M-r" 'flex-isearch-backward)

(use-package iedit
  :bind (("C-;" . iedit-mode)
         ("C-:" . iedit-mode-toggle-on-function)
         (:map iedit-mode-keymap
               ("C-h" . backward-delete-char))))

(with-eval-after-load 'markdown-mode
  (defun cp/copy-md-link (prefix-arg)
    "Copy address of Markdown link after point in the current line.

If there is no link in the current line, or if the region is
active, or with a prefix arg - call
whole-line-or-region-kill-ring-save instead."
    (interactive "P")
    (save-excursion
      (if (or (use-region-p)
              prefix-arg
              (not (cp/re-search-line
                    ;; 2018-03-21T22:47:55+0530 - fix bug where a line with parenthesized text would not be copied
                    ;; "\("
                    ;; 2018-07-22T10:20:03+0530
                    ;; "\\[.*?\\](.*?)"
                    "\\[.*?\\]("
                    )))
          (whole-line-or-region-kill-ring-save prefix-arg)
        (let ((point-a (point)))
          (forward-char -1)
          (forward-sexp)
          (copy-region-as-kill point-a (- (point) 1))))))
  (cp-set-keys
   :keymap markdown-mode-map
   :bindings
   `((,(kbd "M-w") cp/copy-md-link)
     (,(kbd "TAB") markdown-cycle)
     (,(kbd "C-c C-.") markdown-demote)
     (,(kbd "C-c C-,") markdown-promote)
     (,(kbd "C-c C-l") markdown-insert-link))))

(defun cp/copy-bus-entry ()
  "For personal use, when working between Markdown and the OSM wiki."
  (interactive)
  (let ((point-a (region-beginning))
        (point-b (region-end))
        (point-b-line (line-number-at-pos)))
    (query-replace "[ ]" "☐" nil point-a point-b)
    (query-replace "[x]" "☑" nil point-a point-b)
    (query-replace-regexp "^- " "" nil point-a point-b)
    (goto-char (point-min))
    (forward-line (- point-b-line 1))
    (copy-region-as-kill point-a (point-at-bol))))

;; 2018-08-21T03:41:47+0530
(defun cp/copy-md-link (prefix-arg)
  "Copy address of Markdown link after point in the current line.

If there is no link in the current line, or if the region is
active, or with a prefix arg - call
whole-line-or-region-kill-ring-save instead."
  (interactive "P")
  (save-excursion
    (cond
     ((or (use-region-p) prefix-arg)
      (whole-line-or-region-kill-ring-save prefix-arg))
     ((cp/re-search-line "\\[.*?\\](")
      (let ((point-a (point)))
        (forward-char -1)
        (forward-sexp)
        (copy-region-as-kill point-a (- (point) 1))))
     ((cp/re-search-line "http")
      (kill-new (thing-at-point 'url)))
     (t (whole-line-or-region-kill-ring-save prefix-arg)))))

;; 2018-08-28T22:50:43+0530
(defun cp/copy-buffer ()
  (interactive)
  (copy-region-as-kill (point-min) (point-max)))

(use-package multiple-cursors
  :bind (("C-M-l" . #'mc/mark-previous-like-this-symbol)
         ("C-M-;" . #'mc/mark-next-like-this-symbol)
         ("C-M-'" . #'mc/mark-all-symbols-like-this)

         ("C-M-," . #'mc/mark-previous-word-like-this)
         ("C-M-." . #'mc/mark-next-word-like-this)
         ("C-M-/" . #'mc/mark-all-words-like-this)
         ("C-M-<" . #'mc/unmark-previous-like-this)
         ("C-M->" . #'mc/unmark-next-like-this)

         ("C-M-]" . #'mc-hide-unmatched-lines-mode)))

(defun cp/marked-files->markup-links-org (filenames)
  (mapcar (lambda (filename)
            (let ((link-pre  "[[file:")
                  (link-post "][]]\n"))
              (concat link-pre filename link-post)))
          filenames))

(defun cp/marked-files->markup-links-md (filenames)
  (mapcar (lambda (filename)
            (if (member (downcase
                         (file-name-extension filename))
                        image-file-name-extensions)
                (let ((link-pre  "![](")
                      (link-post ")\n"))
                  (concat link-pre filename link-post))
              (let ((link-pre  "[](")
                    (link-post ")\n"))
                (concat link-pre filename link-post))))
          filenames))

;; BUG - (dired-get-marked-files) returns file name at point if none are marked
;; BUG - (dired-get-marked-files) always returns an alphabetically-sorted list, even when the Dired buffer is sorted by date
(defun cp/marked-files->markup-links ()
  "From a Dired buffer, insert the file at point or marked files
as links into an Org or Markdown document."
  (interactive)
  (if (derived-mode-p 'dired-mode)
      (let* ((filenames (dired-get-marked-files 'no-dir))
             (other-buffer  (->> (window-list)
                                 (cadr)
                                 (window-buffer)))
             (other-buffer-mode (with-current-buffer other-buffer major-mode))
             (two-window-op (and (= (length (window-list)) 2)
                                 (member other-buffer-mode
                                         '(org-mode markdown-mode))))
             (output-buffer (if two-window-op
                                other-buffer
                              (read-buffer "Insert links in buffer: ")))
             (output-mode (with-current-buffer output-buffer major-mode))
             (output-text (cl-case output-mode
                            ('org-mode
                             (cp/marked-files->markup-links-org filenames))
                            ('markdown-mode
                             (cp/marked-files->markup-links-md filenames))
                            (t (error "Only Markdown and Org are currently supported.")))))
        (if two-window-op
            (other-window 1)
          (switch-to-buffer output-buffer))
        (mapc #'insert output-text))
    (user-error "Run this command from a Dired buffer with some marked files."))
  (when markdown-inline-image-overlays
    (markdown-display-inline-images)))

(use-package adaptive-wrap
  :hook (markdown-mode . adaptive-wrap-prefix-mode))

(use-package avy
  :config
  (setq avy-case-fold-search nil))

(provide 'cp-editing)
